// https://www.redblobgames.com/grids/hexagons/

// Horizontal distance between hex centers is sqrt(3) * size. The vertical
// distance is 3 / 2 * size. When we calculate horzDist / vertDist, the size
// cancels out, leaving us with a unitless ratio of sqrt(3) / (3 / 2), or
// 2 * sqrt(3) / 3.

const svgns = "http://www.w3.org/2000/svg",
  horzToVertDistRatio = 2 * Math.sqrt(3) / 3,

  arcSize = {
    'small': Math.atan(horzToVertDistRatio / 6),
    'medium': Math.atan(horzToVertDistRatio / 2),
    'large': Math.atan(7 * horzToVertDistRatio / 2)
  },

  firingArcVisibility = {
    davion: false,
    liao: false
  };

let svg;

function calculateAngle(xDiff, yDiff) {
  yDiff = -yDiff;
  let angle = Math.abs(Math.atan(yDiff / xDiff));

  if (xDiff < 0 && yDiff > 0) {
    angle = Math.PI - angle;
  } else if (xDiff < 0 && yDiff < 0) {
    angle = Math.PI + angle;
  } else if (xDiff > 0 && yDiff < 0) {
    angle = 2 * Math.PI - angle;
  }

  return angle;
}

function edgePoint(x1, y1, x2, y2, maxX, maxY, minX = 0, minY = 0) {
  let pointCoords,
    xDiff = x2 - x1,
    yDiff = y2 - y1,
    xIntercept = y => (y - y1) * xDiff / yDiff + x1,
    yIntercept = x => (x - x1) * yDiff / xDiff + y1;

  if (xDiff > 0 && yDiff > 0) {
    let x = xIntercept(maxY);

    pointCoords = x <= maxX ? [x, maxY] : [maxX, yIntercept(maxX)];
  } else if (xDiff > 0 && yDiff < 0) {
    let y = yIntercept(maxX);

    pointCoords = y >= minY ? [maxX, y] : [xIntercept(minY), minY];
  } else if (xDiff < 0 && yDiff < 0) {
    let x = xIntercept(minY);

    pointCoords = x >= minX ? [x, minY] : [minX, yIntercept(minX)];
  } else {
    let y = yIntercept(minX);

    pointCoords = y <= maxY ? [minX, y] : [xIntercept(maxY), maxY];
  }

  return pointCoords;
}

let cornerPoints;

function position(e) {
  // let activeFiringArc = this.querySelector('polygon.firing-arc.active');
  let activeFiringArc;
  let size = this.querySelector('polygon.firing-arc.active').dataset.size;

  // TODO: handle exactly horizontal and vertical lines

  let grid = this.querySelector('.grid');
  let testAim = this.querySelector('#test-arcs line');
  let testOutline = this.querySelector('#test-arcs polyline');
  let testArea = this.querySelector('#test-arcs polygon');

  let pt = new DOMPoint(e.clientX, e.clientY);
  let { x, y } = pt.matrixTransform(grid.getScreenCTM().inverse());

  let pivotPt = [x1, y1] = [testAim.getAttribute('x1'), testAim.getAttribute('y1')].map(n => parseFloat(n));
  let [maxX, maxY, minX, minY] = [
    grid.getBBox().x + grid.getBBox().width,
    grid.getBBox().y + grid.getBBox().height,
    grid.getBBox().x,
    grid.getBBox().y
  ].map(n => parseFloat(n));

  let edgePointArgs = [testAim.getAttribute('x1'), testAim.getAttribute('y1'), x, y, maxX, maxY, minX, minY].map(n => parseFloat(n))

  let aimPt = [x2, y2] = edgePoint(...edgePointArgs);

  testAim.setAttributeNS(null, 'x2', x2);
  testAim.setAttributeNS(null, 'y2', y2);

  // console.log(testAim);


  let angle = calculateAngle(x2 - x1, y2 - y1);

  let arcAngle = arcSize[size];
  // let arcAngle = Math.PI / 4;
  let distance = Math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2);
  let yDelta = distance * Math.cos(angle) * Math.tan(arcAngle);
  let xDelta = distance * Math.sin(angle) * Math.tan(arcAngle);

  let [newY1, newX1] = [y2 - yDelta, x2 - xDelta];
  let [newY2, newX2] = [y2 + yDelta, x2 + xDelta];

  let arcPt1 = [newX1, newY1] = edgePoint(x1, y1, newX1, newY1, maxX, maxY, minX, minY);
  let arcPt2 = [newX2, newY2] = edgePoint(x1, y1, newX2, newY2, maxX, maxY, minX, minY);

  testOutline.setAttributeNS(null, 'points', `${newX1},${newY1} ${x1},${y1} ${newX2},${newY2}`);

  // console.log('area    ', testArea.getAttribute('points').split(' ').map(n => n.split(',').map(n => Math.round(parseFloat(n)))).join('  \t'));
  // console.log('lines   ', testOutline.getAttribute('points').split(' ').map(n => n.split(',').map(n => Math.round(parseFloat(n)))).join('  \t'));

  function touchSameEdge([x1, y1], [x2, y2]) {
    // console.log('touchSameEdge', x1, y1, x2, y2);
    return x1 === x2 || y1 === y2;
  }

  function touchOrthogonalEdges([x1, y1], [x2, y2]) {
    let xLimit = [minX, maxX], yLimit = [minY, maxY];
    return (xLimit.includes(x1) && yLimit.includes(y2)) || (yLimit.includes(y1) && xLimit.includes(x2));
  }

  function shareValue([x1, y1], [x2, y2]) {
    return x1 === x2 || y1 === y2;
  }

  function findSharedValue(pt, ...pts) {
    // which arcpt does aimpt share a value with?

    // if they share an x value, we will look for corner y value
    // if they share a y value, we will look for corner x value
    // is aim pt non-shared value greater or less than arcpt non-shared value?

    const sharedValPt = pts.find(([ptX, ptY]) => pt.includes(ptX) || pt.includes(ptY));
    let limits = [[minX, maxX], [minY, maxY]];

    if (!sharedValPt) {
      return;
    }

    const nonSharedValIndex = pt[0] === sharedValPt[0] ? 1 : 0;
    const cornerVal = pt[nonSharedValIndex] < sharedValPt[nonSharedValIndex] ? Math.min(...limits[nonSharedValIndex]) : Math.max(...limits[nonSharedValIndex]);
    let cps = cornerPoints.filter(cp => cp[nonSharedValIndex] === cornerVal);

    return cps;
  }

  console.log('cornerPoints', cornerPoints);

  if (touchSameEdge(arcPt1, arcPt2)) {
    // 0 corner case
    console.log('0-corner case');
    testArea.setAttributeNS(null, 'points', [pivotPt, arcPt1, arcPt2].join(' '));
  } else if (touchOrthogonalEdges(arcPt1, arcPt2)) {
    // 1 corner or 3 corner case
    if (touchSameEdge(aimPt, arcPt1) || touchSameEdge(aimPt, arcPt2)) {
      console.log('1-corner case');
      let cp = cornerPoints.find(cp => shareValue(cp, arcPt1) && shareValue(cp, arcPt2));
      testArea.setAttributeNS(null, 'points', [pivotPt, arcPt1, cp, arcPt2].join(' '));
    } else {
      console.log('3-corner case');
      console.log('cornerPoints', cornerPoints);
      let cps = cornerPoints.filter(cp => !shareValue(cp, arcPt1) || !shareValue(cp, arcPt2));
      let index = cps.findIndex(cp => shareValue(cp, arcPt2));
      console.log('cps', cps, 'index', index, 'arcPt2', arcPt2, 'arcPt1', arcPt1);
      cps.splice(index + 1, 0, arcPt2, pivotPt, arcPt1);
      console.log('after splice', cps);
      // testArea.setAttributeNS(null, 'points', [pivotPt, arcPt1, cps.join(' '), arcPt2].join(' '));
      testArea.setAttributeNS(null, 'points', cps.join(' '));
    }
  } else {
    // 2 corner case
    if (touchSameEdge(aimPt, arcPt1) || touchSameEdge(aimPt, arcPt2)) {
      console.log('2-corner case, edge shared');
      let cps = findSharedValue(aimPt, arcPt1, arcPt2);
      let index = cps.findIndex(cp => shareValue(cp, arcPt2));
      // console.log('cps', cps, 'index', index, 'arcPt2', arcPt2, 'arcPt1', arcPt1);
      cps.splice(index + 1, 0, arcPt2, pivotPt, arcPt1);

      // testArea.setAttributeNS(null, 'points', [pivotPt, arcPt1, cps.join(' '), arcPt2].join(' '));
      testArea.setAttributeNS(null, 'points', cps.join(' '));
    } else {
      console.log('2-corner case, all separate edges');
      let cps = cornerPoints.filter(cp => shareValue(cp, aimPt) || shareValue(cp, aimPt));
      let index = cps.findIndex(cp => shareValue(cp, arcPt2));
      cps.splice(index + 1, 0, arcPt2, pivotPt, arcPt1);

      // testArea.setAttributeNS(null, 'points', [pivotPt, arcPt1, cps.join(' '), arcPt2].join(' '));
      testArea.setAttributeNS(null, 'points', cps.join(' '));
    }
  }

  // testArea.setAttributeNS(null, 'points', cornerPoints.join(' '));












  if (activeFiringArc) {
    let activeFiringArcOutline = this.querySelector(`#lines polygon[data-number="${activeFiringArc.dataset.number}"][data-allegiance="${activeFiringArc.dataset.allegiance}"]`),
      board = this.querySelector('#image-maps'),
      { width, height } = board.getBBox(),
      pt = new DOMPoint(e.clientX, e.clientY),
      { x: pointerX, y: pointerY } = pt.matrixTransform(board.getScreenCTM().inverse()),
      [maxXpx, maxYpx] = [width, height],
      { x: x1px, y: y1px } = activeFiringArc.points[0];

    let [x2px, y2px] = [
      pointerX / width * maxXpx,
      pointerY / height * maxYpx
    ];

    let xDiff = x2px - x1px;
    let yDiff = y2px - y1px;
    let angle = calculateAngle(xDiff, yDiff);

    let arcAngle = arcSize[activeFiringArc.dataset.size];
    let distance = Math.sqrt((x2px - x1px) ** 2 + (y2px - y1px) ** 2);
    let yDelta = distance * Math.cos(angle) * Math.tan(arcAngle);
    let xDelta = distance * Math.sin(angle) * Math.tan(arcAngle);

    let [newY1, newX1] = [y2px + yDelta, x2px + xDelta];
    let [newY2, newX2] = [y2px - yDelta, x2px - xDelta];

    [newX1, newY1] = edgePoint(x1px, y1px, newX1, newY1, maxXpx, maxYpx);
    [newX2, newY2] = edgePoint(x1px, y1px, newX2, newY2, maxXpx, maxYpx);

    let oppositeEdgeConditions = [
      newX1 == 0 && newX2 == maxXpx,
      newX2 == 0 && newX1 == maxXpx,
      newY1 == 0 && newY2 == maxYpx,
      newY2 == 0 && newY1 == maxYpx
    ]

    let orthogonalEdgeConditions = [
      (newX1 == 0 || newX1 == maxXpx) && (newY2 == 0 || newY2 == maxYpx),
      (newX2 == 0 || newX2 == maxXpx) && (newY1 == 0 || newY1 == maxYpx),
    ]

    let points;

    if (oppositeEdgeConditions.some(e => e)) {
      let cornerPoints;

      if (xDiff > 0 && yDiff > 0) {
        if ((newY1 == 0 && newY2 == maxYpx) || (newY1 == maxYpx && newY2 == 0)) {
          cornerPoints = [[maxXpx, 0], [maxXpx, maxYpx]];
        } else {
          cornerPoints = [[maxXpx, maxYpx], [0, maxYpx]];
        }
      } else if (xDiff > 0 && yDiff < 0) {
        if ((newY1 == 0 && newY2 == maxYpx) || (newY1 == maxYpx && newY2 == 0)) {
          cornerPoints = [[maxXpx, 0], [maxXpx, maxYpx]];
        } else {
          cornerPoints = [[0, 0], [maxXpx, 0]];
        }

      } else if (xDiff < 0 && yDiff > 0) {
        if ((newY1 == 0 && newY2 == maxYpx) || (newY1 == maxYpx && newY2 == 0)) {
          cornerPoints = [[0, maxYpx], [0, 0]];
        } else {
          cornerPoints = [[maxXpx, maxYpx], [0, maxYpx]];
        }

      } else {
        if ((newY1 == 0 && newY2 == maxYpx) || (newY1 == maxYpx && newY2 == 0)) {
          cornerPoints = [[0, maxYpx], [0, 0]];
        } else {
          cornerPoints = [[0, 0], [maxXpx, 0]];
        }

      }
      points = `${x1px},${y1px} ${newX1},${newY1} ${cornerPoints[1]} ${cornerPoints[0]} ${newX2},${newY2}`;
    } else if (orthogonalEdgeConditions.some(e => e)) {
      let cornerPoints = [];
      let cp1, cp2;

      if (newX1 == 0 || newX1 == maxXpx) {
        cp1 = [newX1, yDiff > 0 ? maxYpx : 0];
      } else {
        cp1 = [xDiff > 0 ? maxXpx : 0, newY1];
      }

      if (newX2 == 0 || newX2 == maxXpx) {
        cp2 = [newX2, yDiff > 0 ? maxYpx : 0];
      } else {
        cp2 = [xDiff > 0 ? maxXpx : 0, newY2];
      }

      if (cp1[0] == cp2[0] && cp1[1] == cp2[1]) {
        cornerPoints.push(cp1);
      } else {
        cornerPoints.push(cp1);
        cornerPoints.push([xDiff > 0 ? maxXpx : 0, yDiff > 0 ? maxYpx : 0])
        cornerPoints.push(cp2);
      }

      points = `${x1px},${y1px} ${newX1},${newY1} ${cornerPoints.join(' ')} ${newX2},${newY2}`;
    } else {
      points = `${x1px},${y1px} ${newX1},${newY1} ${newX2},${newY2}`;
    }

    activeFiringArcOutline.setAttributeNS(null, 'points', points);
    activeFiringArc.setAttributeNS(null, 'points', points);
  }
}

function setDataAttrs({ dataset: { allegiance, number }}, el) {
  el.dataset.allegiance = allegiance;
  el.dataset.number = number;
}

function getClipPathId({ dataset: { allegiance, number }}) {
  return `clip-path-${allegiance}-${number}`;
}

function getUnclipped() {
  return svg.querySelectorAll('#firing-arcs polygon:not([clip-path])');
};

export default function (el) {
  svg = el;

  this.set = function (size, counter, { x, y }) {
    this.get(counter).forEach(fa => fa.remove());

    let arcLayer = svg.querySelector('#shapes');
    let outlineLayer = svg.querySelector('#lines');
    let arcContainer = svg.querySelector('#firing-arcs');




    let testArcs = svg.querySelector('#test-arcs');
    let testAim = document.createElementNS(svgns, 'line');
    testAim.setAttributeNS(null, 'x1', x);
    testAim.setAttributeNS(null, 'y1', y);
    testAim.setAttributeNS(null, 'x2', x);
    testAim.setAttributeNS(null, 'y2', y);
    testArcs.appendChild(testAim);

    let testOutline = document.createElementNS(svgns, 'polyline');
    testOutline.setAttributeNS(null, 'points', `${x},${y} ${x},${y} ${x},${y}`);
    testArcs.appendChild(testOutline);

    let testGrid = svg.querySelector('.grid');
    let gridArea = document.createElementNS(svgns, 'polygon');

    cornerPoints = (function () {
      let { x, y, width, height } = testGrid.getBBox();
      return [[x, y], [x + width, y], [x + width, y + height], [x, y + height]];
    })();

    gridArea.setAttributeNS(null, 'points', cornerPoints.join(' '));
    testArcs.appendChild(gridArea);




    // console.log(testArcs);

    let grid = svg.querySelector('.board');
    const transform = getComputedStyle(grid).transform.match(/-?\d+\.?\d*/g);
    const pt = new DOMPoint(x, y);
    const mtx = new DOMMatrix(transform);
    let tPt = pt.matrixTransform(mtx);

    let pivotPoint = [tPt.x, tPt.y];
    let firingArc = document.createElementNS(svgns, 'polygon');
    let firingArcOutline = document.createElementNS(svgns, 'polygon');

    setDataAttrs(counter, firingArc);
    firingArc.dataset.size = size;
    firingArc.classList.add('firing-arc', 'active');
    firingArc.setAttributeNS(null, 'points', `${pivotPoint} ${pivotPoint} ${pivotPoint}`);

    setDataAttrs(counter, firingArcOutline);
    firingArcOutline.setAttributeNS(null, 'points', `${pivotPoint} ${pivotPoint} ${pivotPoint}`);

    let clipShape = document.createElementNS(svgns, 'circle');
    clipShape.setAttributeNS(null, 'cx', tPt.x);
    clipShape.setAttributeNS(null, 'cy', tPt.y);
    clipShape.setAttributeNS(null, 'r', 100);

    let clipPath = document.createElementNS(svgns, 'clipPath');
    setDataAttrs(counter, clipPath);
    clipPath.setAttributeNS(null, 'id', getClipPathId(counter));
    clipPath.appendChild(clipShape);

    arcContainer.appendChild(clipPath);
    arcLayer.appendChild(firingArc);
    outlineLayer.appendChild(firingArcOutline);

    let firingArcPlacementListener = e => {
      svg.querySelectorAll('.firing-arc.active').forEach(el => el.classList.remove('active'));
      grid.removeAttribute('style');
      svg.removeEventListener('mousemove', position);
      firingArc.removeEventListener('click', firingArcPlacementListener);
      firingArc.removeEventListener('contextmenu', cancelFiringArcPlacement);
    };

    let cancelFiringArcPlacement = e => {
      e.preventDefault();

      firingArc.removeEventListener('click', firingArcPlacementListener);
      firingArc.removeEventListener('contextmenu', cancelFiringArcPlacement);

      this.get(counter).forEach(fa => fa.remove());

      grid.removeAttribute('style');
      svg.removeEventListener('mousemove', position);
    };

    grid.style.pointerEvents = 'none';
    svg.addEventListener('mousemove', position);
    firingArc.addEventListener('click', firingArcPlacementListener);
    firingArc.addEventListener('contextmenu', cancelFiringArcPlacement);
  };

  this.clear = function (allegiance) {
    const selector = `#firing-arcs [data-allegiance="${allegiance}"]`;
    svg.querySelectorAll(selector).forEach(el => el.remove());
  };

  this.get = function ({ dataset: { allegiance, number }}) {
    return svg.querySelectorAll(`#firing-arcs polygon[data-number="${number}"][data-allegiance="${allegiance}"]`);
  };

  this.toggleVisibility = function (allegiance) {
    const vis = firingArcVisibility[allegiance],
      clipPaths = svg.querySelectorAll(`clipPath[data-allegiance="${allegiance}"]`);

    clipPaths.forEach(cp => cp.style.display = !vis ? 'none' : '');
    firingArcVisibility[allegiance] = !vis;
  };

  this.toggleCounterVisibility = function ({ dataset: { number, allegiance }}, vis) {
    const cp = svg.querySelector(`#clip-path-${allegiance}-${number}`),
      display = vis ? 'none' : '';

    if (cp) {
      cp.style.display = firingArcVisibility[allegiance] ? 'none' : display;
    }
  };

  this.clipAll = function () {
    console.log('clipall')
    let unclipped = getUnclipped();

    unclipped.forEach(el => {
      const { number, allegiance } = el.dataset,
        clipPathId = `clip-path-${allegiance}-${number}`,
        isVisible = firingArcVisibility[allegiance];

      if (isVisible) {
        svg.querySelector(`#${clipPathId}`).style.display = 'none';
      }

      el.setAttributeNS(null, 'clip-path', `url(#${clipPathId})`);
    });
  };
}
